@@ -19,6 +19,7 @@
 package io.undertow;
 
 import io.undertow.client.ClientConnection;
+import io.undertow.protocols.ssl.SslConduit;
 import io.undertow.server.HttpServerExchange;
 import io.undertow.server.ServerConnection;
 import io.undertow.server.handlers.sse.ServerSentEventConnection;
@@ -30,6 +31,7 @@ import org.jboss.logging.annotations.Cause;
 import org.jboss.logging.annotations.LogMessage;
 import org.jboss.logging.annotations.Message;
 import org.jboss.logging.annotations.MessageLogger;
+import org.xnio.ssl.SslConnection;
 
 import java.io.IOException;
 import java.net.InetAddress;
@@ -58,6 +60,8 @@ public interface UndertowLogger extends BasicLogger {
     UndertowLogger CLIENT_LOGGER = Logger.getMessageLogger(UndertowLogger.class, ClientConnection.class.getPackage().getName());
 
     UndertowLogger REQUEST_LOGGER = Logger.getMessageLogger(UndertowLogger.class, UndertowLogger.class.getPackage().getName() + ".request");
+    UndertowLogger SESSION_LOGGER = Logger.getMessageLogger(UndertowLogger.class, UndertowLogger.class.getPackage().getName() + ".session");
+    UndertowLogger SECURITY_LOGGER = Logger.getMessageLogger(UndertowLogger.class, UndertowLogger.class.getPackage().getName() + ".request.security");
     UndertowLogger PROXY_REQUEST_LOGGER = Logger.getMessageLogger(UndertowLogger.class, UndertowLogger.class.getPackage().getName() + ".proxy");
     UndertowLogger REQUEST_DUMPER_LOGGER = Logger.getMessageLogger(UndertowLogger.class, UndertowLogger.class.getPackage().getName() + ".request.dump");
     /**
@@ -66,23 +70,23 @@ public interface UndertowLogger extends BasicLogger {
      */
     UndertowLogger REQUEST_IO_LOGGER = Logger.getMessageLogger(UndertowLogger.class, UndertowLogger.class.getPackage().getName() + ".request.io");
 
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = ERROR)
     @Message(id = 5001, value = "An exception occurred processing the request")
     void exceptionProcessingRequest(@Cause Throwable cause);
 
-    @LogMessage(level = INFO)
-    @Message(id = 5002, value = "Exception reading file %s: %s")
-    void exceptionReadingFile(final Path file, final IOException e);
+//    @LogMessage(level = INFO)
+//    @Message(id = 5002, value = "Exception reading file %s: %s")
+//    void exceptionReadingFile(final Path file, final IOException e);
 
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = ERROR)
     @Message(id = 5003, value = "IOException reading from channel")
     void ioExceptionReadingFromChannel(@Cause IOException e);
 
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = ERROR)
     @Message(id = 5005, value = "Cannot remove uploaded file %s")
     void cannotRemoveUploadedFile(Path file);
 
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = DEBUG)
     @Message(id = 5006, value = "Connection from %s terminated as request header was larger than %s")
     void requestHeaderWasTooLarge(SocketAddress address, int size);
 
@@ -102,7 +106,7 @@ public interface UndertowLogger extends BasicLogger {
     @Message(id = 5010, value = "Verification of authentication tokens for user '%s' has failed using mechanism '%s'.")
     void authenticationFailed(final String userName, final String mechanism);
 
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = ERROR)
     @Message(id = 5011, value = "Ignoring AJP request with prefix %s")
     void ignoringAjpRequestWithPrefixCode(byte prefix);
 
@@ -111,82 +115,82 @@ public interface UndertowLogger extends BasicLogger {
     void ioException(@Cause IOException e);
 
     @LogMessage(level = DEBUG)
-    @Message(id = 5014, value = "Failed to parse HTTP request")
-    void failedToParseRequest(@Cause Exception e);
+    @Message(id = 5014, value = "Failed to parse request")
+    void failedToParseRequest(@Cause Throwable e);
 
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = ERROR)
     @Message(id = 5015, value = "Error rotating access log")
     void errorRotatingAccessLog(@Cause IOException e);
 
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = ERROR)
     @Message(id = 5016, value = "Error writing access log")
     void errorWritingAccessLog(@Cause IOException e);
 
-    @LogMessage(level = Logger.Level.ERROR)
-    @Message(id = 5017, value = "Unknown variable %s")
+    @LogMessage(level = ERROR)
+    @Message(id = 5017, value = "Unknown variable %s. For the literal percent character use two percent characters: '%%'")
     void unknownVariable(String token);
 
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = ERROR)
     @Message(id = 5018, value = "Exception invoking close listener %s")
     void exceptionInvokingCloseListener(ServerConnection.CloseListener l, @Cause Throwable e);
 
-    @LogMessage(level = Logger.Level.ERROR)
-    @Message(id = 5019, value = "Cannot upgrade connection")
-    void cannotUpgradeConnection(@Cause Exception e);
+//    @LogMessage(level = Logger.Level.ERROR)
+//    @Message(id = 5019, value = "Cannot upgrade connection")
+//    void cannotUpgradeConnection(@Cause Exception e);
 
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = ERROR)
     @Message(id = 5020, value = "Error writing JDBC log")
     void errorWritingJDBCLog(@Cause SQLException e);
 
-    @LogMessage(level = Logger.Level.ERROR)
-    @Message(id = 5021, value = "Proxy request to %s timed out")
-    void proxyRequestTimedOut(String requestURI);
+//    @LogMessage(level = Logger.Level.ERROR)
+//    @Message(id = 5021, value = "Proxy request to %s timed out")
+//    void proxyRequestTimedOut(String requestURI);
 
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = ERROR)
     @Message(id = 5022, value = "Exception generating error page %s")
     void exceptionGeneratingErrorPage(@Cause Exception e, String location);
 
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = ERROR)
     @Message(id = 5023, value = "Exception handling request to %s")
     void exceptionHandlingRequest(@Cause Throwable t, String requestURI);
 
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = ERROR)
     @Message(id = 5024, value = "Could not register resource change listener for caching resource manager, automatic invalidation of cached resource will not work")
     void couldNotRegisterChangeListener(@Cause Exception e);
 
-    @LogMessage(level = Logger.Level.ERROR)
-    @Message(id = 5025, value = "Could not initiate SPDY connection and no HTTP fallback defined")
-    void couldNotInitiateSpdyConnection();
+//    @LogMessage(level = Logger.Level.ERROR)
+//    @Message(id = 5025, value = "Could not initiate SPDY connection and no HTTP fallback defined")
+//    void couldNotInitiateSpdyConnection();
+//
+//    @LogMessage(level = INFO)
+//    @Message(id = 5026, value = "Jetty ALPN support not found on boot class path, %s client will not be available.")
+//    void jettyALPNNotFound(String protocol);
 
-    @LogMessage(level = Logger.Level.ERROR)
-    @Message(id = 5026, value = "Jetty ALPN support not found on boot class path, %s client will not be available.")
-    void jettyALPNNotFound(String protocol);
-
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = ERROR)
     @Message(id = 5027, value = "Timing out request to %s")
     void timingOutRequest(String requestURI);
 
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = ERROR)
     @Message(id = 5028, value = "Proxy request to %s failed")
     void proxyRequestFailed(String requestURI, @Cause Exception e);
 
-    @LogMessage(level = Logger.Level.ERROR)
-    @Message(id = 5030, value = "Proxy request to %s could not resolve a backend server")
-    void proxyRequestFailedToResolveBackend(String requestURI);
+//    @LogMessage(level = Logger.Level.ERROR)
+//    @Message(id = 5030, value = "Proxy request to %s could not resolve a backend server")
+//    void proxyRequestFailedToResolveBackend(String requestURI);
 
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = ERROR)
     @Message(id = 5031, value = "Proxy request to %s could not connect to backend server %s")
     void proxyFailedToConnectToBackend(String requestURI, URI uri);
 
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = ERROR)
     @Message(id = 5032, value = "Listener not making progress on framed channel, closing channel to prevent infinite loop")
     void listenerNotProgressing();
 
-    @LogMessage(level = Logger.Level.ERROR)
-    @Message(id = 5033, value = "Failed to initiate HTTP2 connection")
-    void couldNotInitiateHttp2Connection();
+//    @LogMessage(level = Logger.Level.ERROR)
+//    @Message(id = 5033, value = "Failed to initiate HTTP2 connection")
+//    void couldNotInitiateHttp2Connection();
 
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = ERROR)
     @Message(id = 5034, value = "Remote endpoint failed to send initial settings frame in HTTP2 connection, frame type %s")
     void remoteEndpointFailedToSendInitialSettings(int type);
 
@@ -194,7 +198,7 @@ public interface UndertowLogger extends BasicLogger {
     @Message(id = 5035, value = "Closing channel because of parse timeout for remote address %s")
     void parseRequestTimedOut(java.net.SocketAddress remoteAddress);
 
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = ERROR)
     @Message(id = 5036, value = "ALPN negotiation failed for %s and no fallback defined, closing connection")
     void noALPNFallback(SocketAddress address);
 
@@ -218,7 +222,7 @@ public interface UndertowLogger extends BasicLogger {
     @Message(id = 5040, value = "Gonna send payload:\n%s")
     void proxyAdvertiseMessagePayload(String payload);
 
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = ERROR)
     @Message(id = 5041, value = "Cannot send advertise message. Address: %s")
     void proxyAdvertiseCannotSendMessage(@Cause Exception e, InetSocketAddress address);
 
@@ -226,7 +230,7 @@ public interface UndertowLogger extends BasicLogger {
     @Message(id = 5042, value = "Undertow mod_cluster proxy MCMPHandler created")
     void mcmpHandlerCreated();
 
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = ERROR)
     @Message(id = 5043, value = "Error in processing MCMP commands: Type:%s, Mess: %s")
     void mcmpProcessingError(String type, String errString);
 
@@ -258,11 +262,11 @@ public interface UndertowLogger extends BasicLogger {
     void nodeConfigCreated(URI connectionURI, String balancer, String domain, String jvmRoute, boolean flushPackets, int flushwait, int ping, long ttl,
                            int timeout, int maxConnections, int cacheConnections, int requestQueueSize, boolean queueNewRequests);
 
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = ERROR)
     @Message(id = 5050, value = "Failed to process management request")
     void failedToProcessManagementReq(@Cause Exception e);
 
-    @LogMessage(level = Logger.Level.ERROR)
+    @LogMessage(level = ERROR)
     @Message(id = 5051, value = "Failed to send ping response")
     void failedToSendPingResponse(@Cause Exception e);
 
@@ -294,7 +298,7 @@ public interface UndertowLogger extends BasicLogger {
     @Message(id = 5058, value = "Could not bind multicast socket to %s (%s address): %s; make sure your multicast address is of the same type as the IP stack (IPv4 or IPv6). Multicast socket will not be bound to an address, but this may lead to cross talking (see http://www.jboss.org/community/docs/DOC-9469 for details).")
     void potentialCrossTalking(InetAddress group, String s, String localizedMessage);
 
-    @LogMessage(level = org.jboss.logging.Logger.Level.WARN)
+    @LogMessage(level = WARN)
     @Message(id = 5060, value = "Predicate %s uses old style square braces to define predicates, which will be removed in a future release. predicate[value] should be changed to predicate(value)")
     void oldStylePredicateSyntax(String string);
 
@@ -352,4 +356,71 @@ public interface UndertowLogger extends BasicLogger {
     @LogMessage(level = ERROR)
     @Message(id = 5074, value = "Failed to invoke error callback %s for SSE task")
     void failedToInvokeFailedCallback(ServerSentEventConnection.EventCallback callback, @Cause Exception e);
+
+    @Message(id = 5075, value = "Unable to resolve mod_cluster management host's address for '%s'")
+    IllegalStateException unableToResolveModClusterManagementHost(String providedHost);
+
+    @LogMessage(level = ERROR)
+    @Message(id = 5076, value = "SSL read loop detected. This should not happen, please report this to the Undertow developers. Current state %s")
+    void sslReadLoopDetected(SslConduit sslConduit);
+
+    @LogMessage(level = ERROR)
+    @Message(id = 5077, value = "SSL unwrap buffer overflow detected. This should not happen, please report this to the Undertow developers. Current state %s")
+    void sslBufferOverflow(SslConduit sslConduit);
+
+//    @LogMessage(level = ERROR)
+//    @Message(id = 5078, value = "ALPN connection failed")
+//    void alpnConnectionFailed(@Cause Exception e);
+
+    @LogMessage(level = ERROR)
+    @Message(id = 5079, value = "ALPN negotiation on %s failed")
+    void alpnConnectionFailed(SslConnection connection);
+
+    @LogMessage(level = ERROR)
+    @Message(id = 5080, value = "HttpServerExchange cannot have both async IO resumed and dispatch() called in the same cycle")
+    void resumedAndDispatched();
+
+    @LogMessage(level = ERROR)
+    @Message(id = 5081, value = "Response has already been started, cannot proxy request %s")
+    void cannotProxyStartedRequest(HttpServerExchange exchange);
+
+    @Message(id = 5082, value = "Configured mod_cluster management host address cannot be a wildcard address (%s)!")
+    IllegalArgumentException cannotUseWildcardAddressAsModClusterManagementHost(String providedAddress);
+
+    @Message(id = 5083, value = "Unexpected end of compressed input")
+    IOException unexpectedEndOfCompressedInput();
+
+    @Message(id = 5084, value = "Attempted to write %s bytes however content-length has been set to %s")
+    IOException dataLargerThanContentLength(long totalToWrite, long responseContentLength);
+
+    @LogMessage(level = ERROR)
+    @Message(id = 5085, value = "Connection %s for exchange %s was not closed cleanly, forcibly closing connection")
+    void responseWasNotTerminated(ServerConnection connection, HttpServerExchange exchange);
+
+    @LogMessage(level = ERROR)
+    @Message(id = 5086, value = "Failed to accept SSL request")
+    void failedToAcceptSSLRequest(@Cause Exception e);
+
+    @LogMessage(level = ERROR)
+    @Message(id = 5087, value = "Failed to use the server order")
+    void failedToUseServerOrder(@Cause ReflectiveOperationException e);
+
+    @LogMessage(level = ERROR)
+    @Message(id = 5088, value = "Failed to execute ServletOutputStream.closeAsync() on IO thread")
+    void closeAsyncFailed(@Cause IOException e);
+
+    @Message(id = 5089, value = "Method parameter '%s' cannot be null")
+    IllegalArgumentException nullParameter(String name);
+
+    @LogMessage(level = ERROR)
+    @Message(id = 5090, value = "Unexpected failure")
+    void handleUnexpectedFailure(@Cause Throwable t);
+
+    @LogMessage(level = ERROR)
+    @Message(id = 5091, value = "Failed to initialize DirectByteBufferDeallocator")
+    void directBufferDeallocatorInitializationFailed(@Cause Throwable t);
+
+    @LogMessage(level = DEBUG)
+    @Message(id = 5092, value = "Failed to free direct buffer")
+    void directBufferDeallocationFailed(@Cause Throwable t);
 }
