@@ -35,6 +35,7 @@ import io.undertow.conduits.BytesSentStreamSinkConduit;
 import io.undertow.conduits.ChunkedStreamSinkConduit;
 import io.undertow.conduits.ChunkedStreamSourceConduit;
 import io.undertow.conduits.ConduitListener;
+import io.undertow.conduits.FinishableStreamSourceConduit;
 import io.undertow.conduits.FixedLengthStreamSourceConduit;
 import io.undertow.protocols.http2.Http2Channel;
 import io.undertow.server.Connectors;
@@ -47,6 +48,7 @@ import io.undertow.util.Methods;
 import io.undertow.util.PooledAdaptor;
 import io.undertow.util.Protocols;
 import io.undertow.util.StatusCodes;
+import org.jboss.logging.Logger;
 import org.xnio.ChannelExceptionHandler;
 import org.xnio.ChannelListener;
 import org.xnio.ChannelListeners;
@@ -69,9 +71,12 @@ import java.io.Closeable;
 import java.io.IOException;
 import java.net.SocketAddress;
 import java.nio.ByteBuffer;
+import java.nio.channels.ClosedChannelException;
 import java.util.ArrayDeque;
 import java.util.Deque;
+import java.util.List;
 import java.util.Locale;
+import java.util.concurrent.CopyOnWriteArrayList;
 
 import static io.undertow.client.UndertowClientMessages.MESSAGES;
 import static io.undertow.util.Headers.CLOSE;
@@ -102,6 +107,8 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
         }
     };
 
+    private static final Logger log = Logger.getLogger(HttpClientConnection.class);
+
     private final Deque<HttpClientExchange> pendingQueue = new ArrayDeque<>();
     private HttpClientExchange currentRequest;
     private HttpResponseBuilder pendingResponse;
@@ -119,7 +126,6 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
     private static final int UPGRADE_REQUESTED = 1 << 29;
     private static final int CLOSE_REQ = 1 << 30;
     private static final int CLOSED = 1 << 31;
-    private int count = 0;
 
     private int state;
     private final ChannelListener.SimpleSetter<HttpClientConnection> closeSetter = new ChannelListener.SimpleSetter<>();
@@ -134,6 +140,7 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
      * The actual connection if this has been upgraded to h2c
      */
     private ClientConnection http2Delegate;
+    private final List<ChannelListener<ClientConnection>> closeListeners = new CopyOnWriteArrayList<>();
 
     HttpClientConnection(final StreamConnection connection, final OptionMap options, final ByteBufferPool bufferPool) {
 
@@ -165,6 +172,7 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
         connection.getCloseSetter().set(new ChannelListener<StreamConnection>() {
 
             public void handleEvent(StreamConnection channel) {
+                log.debugf("connection to %s closed", getPeerAddress());
                 HttpClientConnection.this.state |= CLOSED;
                 ChannelListeners.invokeChannelListener(HttpClientConnection.this, closeSetter.get());
                 try {
@@ -172,8 +180,25 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
                         pooledBuffer.close();
                     }
                 } catch (Throwable ignored){}
+
+                for(ChannelListener<ClientConnection> listener : closeListeners) {
+                    listener.handleEvent(HttpClientConnection.this);
+                }
+                HttpClientExchange pending = pendingQueue.poll();
+                while (pending != null) {
+                    pending.setFailed(new ClosedChannelException());
+                    pending = pendingQueue.poll();
+                }
+                if(currentRequest != null) {
+                    currentRequest.setFailed(new ClosedChannelException());
+                    currentRequest = null;
+                    pendingResponse = null;
+                }
             }
         });
+        //we resume reads, so if the target goes away we get notified
+        connection.getSourceChannel().setReadListener(clientReadListener);
+        connection.getSourceChannel().resumeReads();
     }
 
     @Override
@@ -294,20 +319,24 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
         return true;
     }
 
+    @Override
+    public void addCloseListener(ChannelListener<ClientConnection> listener) {
+        closeListeners.add(listener);
+    }
+
     @Override
     public void sendRequest(final ClientRequest request, final ClientCallback<ClientExchange> clientCallback) {
         if(http2Delegate != null) {
             http2Delegate.sendRequest(request, clientCallback);
             return;
         }
-        count++;
         if (anyAreSet(state, UPGRADE_REQUESTED | UPGRADED | CLOSE_REQ | CLOSED)) {
             clientCallback.failed(UndertowClientMessages.MESSAGES.invalidConnectionState());
             return;
         }
         final HttpClientExchange httpClientExchange = new HttpClientExchange(clientCallback, request, this);
         boolean ssl = this.connection instanceof SslConnection;
-        if(!ssl && !http2Tried && options.get(UndertowOptions.ENABLE_HTTP2, false) && !request.getRequestHeaders().contains(Headers.UPGRADE) && request.getMethod().equals(Methods.GET)) {
+        if(!ssl && !http2Tried && options.get(UndertowOptions.ENABLE_HTTP2, false) && !request.getRequestHeaders().contains(Headers.UPGRADE)) {
             //this is the first request, as we want to try a HTTP2 upgrade
             request.getRequestHeaders().put(new HttpString("HTTP2-Settings"), Http2ClearClientProvider.createSettingsFrame(options, bufferPool));
             request.getRequestHeaders().put(Headers.UPGRADE, Http2Channel.CLEARTEXT_UPGRADE_STRING);
@@ -354,7 +383,9 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
 
         ConduitStreamSinkChannel sinkChannel = connection.getSinkChannel();
         StreamSinkConduit conduit = originalSinkConduit;
-        conduit = new HttpRequestConduit(conduit, bufferPool, request);
+        HttpRequestConduit httpRequestConduit = new HttpRequestConduit(conduit, bufferPool, request);
+        httpClientExchange.setRequestConduit(httpRequestConduit);
+        conduit = httpRequestConduit;
 
         String fixedLengthString = request.getRequestHeaders().getFirst(CONTENT_LENGTH);
         String transferEncodingString = request.getRequestHeaders().getLast(TRANSFER_ENCODING);
@@ -367,7 +398,7 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
                 conduit = new ClientFixedLengthStreamSinkConduit(conduit, length, false, false, currentRequest);
                 hasContent = length != 0;
             } catch (NumberFormatException e) {
-                handleError(new IOException(e));
+                handleError(e);
                 return;
             }
         } else if (transferEncodingString != null) {
@@ -397,19 +428,27 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
                     }));
                     sinkChannel.resumeWrites();
                 }
-            } catch (IOException e) {
-                handleError(e);
+            } catch (Throwable t) {
+                handleError(t);
             }
         }
     }
 
+    private void handleError(Throwable exception) {
+        if (exception instanceof IOException) {
+            handleError((IOException) exception);
+        } else {
+            handleError(new IOException(exception));
+        }
+    }
     private void handleError(IOException exception) {
-        currentRequest.setFailed(exception);
         UndertowLogger.REQUEST_IO_LOGGER.ioException(exception);
         safeClose(connection);
+        currentRequest.setFailed(exception);
     }
 
     public StreamConnection performUpgrade() throws IOException {
+        log.debugf("connection to %s is being upgraded", getPeerAddress());
         // Upgrade the connection
         // Set the upgraded flag already to prevent new requests after this one
         if (allAreSet(state, UPGRADED | CLOSE_REQ | CLOSED)) {
@@ -422,6 +461,7 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
     }
 
     public void close() throws IOException {
+        log.debugf("close called on connection to %s", getPeerAddress());
         if(http2Delegate != null) {
             http2Delegate.close();
         }
@@ -436,6 +476,7 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
      * Notification that the current request is finished
      */
     public void exchangeDone() {
+        log.debugf("exchange complete in connection to %s", getPeerAddress());
 
         connection.getSinkChannel().setConduit(originalSinkConduit);
         connection.getSourceChannel().setConduit(pushBackStreamSourceConduit);
@@ -444,17 +485,19 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
 
         if (anyAreSet(state, CLOSE_REQ)) {
             currentRequest = null;
+            pendingResponse = null;
             this.state |= CLOSED;
             safeClose(connection);
         } else if (anyAreSet(state, UPGRADE_REQUESTED)) {
             connection.getSourceChannel().suspendReads();
             currentRequest = null;
+            pendingResponse = null;
             return;
         }
         currentRequest = null;
+        pendingResponse = null;
 
         HttpClientExchange next = pendingQueue.poll();
-
         if (next == null) {
             //we resume reads, so if the target goes away we get notified
             connection.getSourceChannel().setReadListener(clientReadListener);
@@ -512,8 +555,13 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
                         if (UndertowLogger.CLIENT_LOGGER.isDebugEnabled()) {
                             UndertowLogger.CLIENT_LOGGER.debugf(e, "Connection closed with IOException");
                         }
-                        safeClose(channel);
-                        currentRequest.setFailed(new IOException(MESSAGES.connectionClosed()));
+                        try {
+                            currentRequest.setFailed(e);
+                            currentRequest = null;
+                            pendingResponse = null;
+                        } finally {
+                            safeClose(channel, HttpClientConnection.this);
+                        }
                         return;
                     }
 
@@ -525,9 +573,14 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
                         return;
                     } else if (res == -1) {
                         channel.suspendReads();
-                        safeClose(HttpClientConnection.this);
-                        // Cancel the current active request
-                        currentRequest.setFailed(new IOException(MESSAGES.connectionClosed()));
+                        try {
+                            // Cancel the current active request
+                            currentRequest.setFailed(new IOException(MESSAGES.connectionClosed()));
+                            currentRequest = null;
+                            pendingResponse = null;
+                        } finally {
+                            safeClose(HttpClientConnection.this);
+                        }
                         return;
                     }
 
@@ -554,10 +607,26 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
                         }
                     }
                 }
-
+                boolean close = false;
                 if(connectionString != null) {
-                    if (HttpString.tryFromString(connectionString).equals(Headers.CLOSE)) {
-                        HttpClientConnection.this.state |= CLOSE_REQ;
+                    HttpString con = new HttpString(connectionString);
+                    if (Headers.CLOSE.equals(con)) {
+                        close = true;
+                    } else if(!response.getProtocol().equals(Protocols.HTTP_1_1)) {
+                        if(!Headers.KEEP_ALIVE.equals(con)) {
+                            close = true;
+                        }
+                    }
+                } else if(!response.getProtocol().equals(Protocols.HTTP_1_1)) {
+                    close = true;
+                }
+                if(close) {
+                    HttpClientConnection.this.state |= CLOSE_REQ;
+                    //we are going to close, kill any queued connections
+                    HttpClientExchange ex = pendingQueue.poll();
+                    while (ex != null) {
+                        ex.setFailed(new IOException(UndertowClientMessages.MESSAGES.connectionClosed()));
+                        ex = pendingQueue.poll();
                     }
                 }
                 if(response.getResponseCode() == StatusCodes.SWITCHING_PROTOCOLS && Http2Channel.CLEARTEXT_UPGRADE_STRING.equals(response.getResponseHeaders().getFirst(Headers.UPGRADE))) {
@@ -591,10 +660,12 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
                 }
 
 
-            } catch (Exception e) {
-                UndertowLogger.CLIENT_LOGGER.exceptionProcessingRequest(e);
+            } catch (Throwable t) {
+                UndertowLogger.CLIENT_LOGGER.exceptionProcessingRequest(t);
                 safeClose(connection);
-                currentRequest.setFailed(new IOException(e));
+                if(currentRequest != null) {
+                    currentRequest.setFailed(new IOException(t));
+                }
             } finally {
                 if (free) {
                     pooled.close();
@@ -610,8 +681,9 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
 
     protected void doHttp2Upgrade() {
         try {
-            Http2Channel http2Channel = new Http2Channel(this.performUpgrade(), null, bufferPool, null, true, true, options);
-            Http2ClientConnection http2ClientConnection = new Http2ClientConnection(http2Channel, currentRequest.getResponseCallback(), currentRequest.getRequest(), currentRequest.getRequest().getRequestHeaders().getFirst(Headers.HOST), clientStatistics);
+            StreamConnection connectedStreamChannel = this.performUpgrade();
+            Http2Channel http2Channel = new Http2Channel(connectedStreamChannel, null, bufferPool, null, true, true, options);
+            Http2ClientConnection http2ClientConnection = new Http2ClientConnection(http2Channel, currentRequest.getResponseCallback(), currentRequest.getRequest(), currentRequest.getRequest().getRequestHeaders().getFirst(Headers.HOST), clientStatistics, false);
             http2ClientConnection.getCloseSetter().set(new ChannelListener<ClientConnection>() {
                 @Override
                 public void handleEvent(ClientConnection channel) {
@@ -619,7 +691,9 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
                 }
             });
             http2Delegate = http2ClientConnection;
+            connectedStreamChannel.getSourceChannel().wakeupReads(); //make sure the read listener is immediately invoked, as it may not happen if data is pushed back
             currentRequest = null;
+            pendingResponse = null;
         } catch (IOException e) {
             UndertowLogger.REQUEST_IO_LOGGER.ioException(e);
             safeClose(this);
@@ -633,18 +707,19 @@ class HttpClientConnection extends AbstractAttachable implements Closeable, Clie
         if (exchange.getRequest().getMethod().equals(Methods.HEAD)) {
             connection.getSourceChannel().setConduit(new FixedLengthStreamSourceConduit(connection.getSourceChannel().getConduit(), 0, responseFinishedListener));
         } else if (chunked) {
-            connection.getSourceChannel().setConduit(new ChunkedStreamSourceConduit(connection.getSourceChannel().getConduit(), pushBackStreamSourceConduit, bufferPool, responseFinishedListener, exchange));
+            connection.getSourceChannel().setConduit(new ChunkedStreamSourceConduit(connection.getSourceChannel().getConduit(), pushBackStreamSourceConduit, bufferPool, responseFinishedListener, exchange, connection));
         } else if (length != null) {
             try {
                 long contentLength = Long.parseLong(length);
                 connection.getSourceChannel().setConduit(new FixedLengthStreamSourceConduit(connection.getSourceChannel().getConduit(), contentLength, responseFinishedListener));
             } catch (NumberFormatException e) {
-                handleError(new IOException(e));
+                handleError(e);
                 throw e;
             }
         } else if (response.getProtocol().equals(Protocols.HTTP_1_1) && !Connectors.isEntityBodyAllowed(response.getResponseCode())) {
             connection.getSourceChannel().setConduit(new FixedLengthStreamSourceConduit(connection.getSourceChannel().getConduit(), 0, responseFinishedListener));
         } else {
+            connection.getSourceChannel().setConduit(new FinishableStreamSourceConduit(connection.getSourceChannel().getConduit(), responseFinishedListener));
             state |= CLOSE_REQ;
         }
     }
