@@ -18,6 +18,7 @@
 
 package io.undertow.protocols.http2;
 
+import static io.undertow.protocols.http2.Http2Channel.DATA_FLAG_END_STREAM;
 import static io.undertow.protocols.http2.Http2Channel.FRAME_TYPE_CONTINUATION;
 import static io.undertow.protocols.http2.Http2Channel.FRAME_TYPE_DATA;
 import static io.undertow.protocols.http2.Http2Channel.FRAME_TYPE_GOAWAY;
@@ -29,11 +30,13 @@ import static io.undertow.protocols.http2.Http2Channel.FRAME_TYPE_SETTINGS;
 import static io.undertow.protocols.http2.Http2Channel.FRAME_TYPE_WINDOW_UPDATE;
 import static io.undertow.protocols.http2.Http2Channel.HEADERS_FLAG_END_HEADERS;
 import static org.xnio.Bits.allAreClear;
+import static org.xnio.Bits.anyAreClear;
 import static org.xnio.Bits.anyAreSet;
 
 import java.io.IOException;
 import java.nio.ByteBuffer;
 
+import io.undertow.UndertowLogger;
 import io.undertow.UndertowMessages;
 import io.undertow.server.protocol.framed.AbstractFramedStreamSourceChannel;
 import io.undertow.server.protocol.framed.FrameHeaderData;
@@ -82,7 +85,7 @@ class Http2FrameHeaderParser implements FrameHeaderData {
                     if (streamId == 0) {
                         throw new ConnectionErrorException(Http2Channel.ERROR_PROTOCOL_ERROR, UndertowMessages.MESSAGES.streamIdMustNotBeZeroForFrameType(Http2Channel.FRAME_TYPE_HEADERS));
                     }
-                    parser = new Http2HeadersParser(length, http2Channel.getDecoder(), http2Channel.isClient(), streamId);
+                    parser = new Http2HeadersParser(length, http2Channel.getDecoder(), http2Channel.isClient(), http2Channel.getMaxHeaders(), streamId, http2Channel.getMaxHeaderListSize());
                     if(allAreClear(flags, Http2Channel.HEADERS_FLAG_END_HEADERS)) {
                         continuationParser = (Http2HeadersParser) parser;
                     }
@@ -109,7 +112,7 @@ class Http2FrameHeaderParser implements FrameHeaderData {
                     break;
                 }
                 case FRAME_TYPE_PUSH_PROMISE: {
-                    parser = new Http2PushPromiseParser(length, http2Channel.getDecoder(), http2Channel.isClient(), streamId);
+                    parser = new Http2PushPromiseParser(length, http2Channel.getDecoder(), http2Channel.isClient(), http2Channel.getMaxHeaders(), streamId, http2Channel.getMaxHeaderListSize());
                     if(allAreClear(flags, Http2Channel.HEADERS_FLAG_END_HEADERS)) {
                         continuationParser = (Http2HeadersParser) parser;
                     }
@@ -203,22 +206,46 @@ class Http2FrameHeaderParser implements FrameHeaderData {
         return length;
     }
 
+    int getActualLength() {
+        return length;
+    }
+
     @Override
     public AbstractFramedStreamSourceChannel<?, ?, ?> getExistingChannel() {
+        Http2StreamSourceChannel http2StreamSourceChannel;
         if (type == FRAME_TYPE_DATA ||
                 type == Http2Channel.FRAME_TYPE_CONTINUATION ||
-                type == Http2Channel.FRAME_TYPE_PRIORITY) {
+                type == Http2Channel.FRAME_TYPE_PRIORITY ) {
             if (anyAreSet(flags, Http2Channel.DATA_FLAG_END_STREAM)) {
-                return http2Channel.removeStreamSource(streamId);
+                http2StreamSourceChannel = http2Channel.removeStreamSource(streamId);
             } else if (type == FRAME_TYPE_CONTINUATION) {
-                Http2StreamSourceChannel channel = http2Channel.getIncomingStream(streamId);
-                if(channel != null && channel.isHeadersEndStream() && anyAreSet(flags, Http2Channel.CONTINUATION_FLAG_END_HEADERS)) {
+                http2StreamSourceChannel = http2Channel.getIncomingStream(streamId);
+                if(http2StreamSourceChannel != null && http2StreamSourceChannel.isHeadersEndStream() && anyAreSet(flags, Http2Channel.CONTINUATION_FLAG_END_HEADERS)) {
                     http2Channel.removeStreamSource(streamId);
                 }
-                return channel;
             } else {
-                return http2Channel.getIncomingStream(streamId);
+                http2StreamSourceChannel = http2Channel.getIncomingStream(streamId);
             }
+            if(type == FRAME_TYPE_DATA && http2StreamSourceChannel != null) {
+                Http2DataFrameParser dataFrameParser = (Http2DataFrameParser) parser;
+                http2StreamSourceChannel.updateContentSize(getFrameLength() - dataFrameParser.getPadding(), anyAreSet(flags, DATA_FLAG_END_STREAM));
+            }
+            return http2StreamSourceChannel;
+        } else if(type == FRAME_TYPE_HEADERS) {
+            //headers can actually be a trailer
+
+            Http2StreamSourceChannel channel = http2Channel.getIncomingStream(streamId);
+            if(channel != null) {
+                if(anyAreClear(flags, Http2Channel.HEADERS_FLAG_END_STREAM)) {
+                    //this is a protocol error
+                    UndertowLogger.REQUEST_IO_LOGGER.debug("Received HTTP/2 trailers header without end stream set");
+                    http2Channel.sendGoAway(Http2Channel.ERROR_PROTOCOL_ERROR);
+                }
+                if (channel.isHeadersEndStream() && anyAreSet(flags, Http2Channel.HEADERS_FLAG_END_HEADERS)) {
+                    http2Channel.removeStreamSource(streamId);
+                }
+            }
+            return channel;
         }
         return null;
     }
