@@ -45,6 +45,7 @@ import io.undertow.util.HttpString;
 import io.undertow.util.Methods;
 import io.undertow.util.NetworkUtils;
 import io.undertow.util.Protocols;
+import io.undertow.util.Rfc6265CookieSupport;
 import io.undertow.util.StatusCodes;
 import org.jboss.logging.Logger;
 import org.xnio.Buffers;
@@ -582,12 +583,11 @@ public final class HttpServerExchange extends AbstractAttachable {
 
     private String extractCharset(HeaderMap headers) {
         String contentType = headers.getFirst(Headers.CONTENT_TYPE);
-        if (contentType == null) {
-            return null;
-        }
-        String value = Headers.extractQuotedValueFromHeader(contentType, "charset");
-        if(value != null) {
-            return value;
+        if (contentType != null) {
+            String value = Headers.extractQuotedValueFromHeader(contentType, "charset");
+            if (value != null) {
+                return value;
+            }
         }
         return ISO_8859_1;
     }
@@ -656,14 +656,16 @@ public final class HttpServerExchange extends AbstractAttachable {
                 colonIndex = host.indexOf(':');
             }
             if (colonIndex != -1) {
-                return Integer.parseInt(host.substring(colonIndex + 1));
-            } else {
-                if (getRequestScheme().equals("https")) {
-                    return 443;
-                } else if (getRequestScheme().equals("http")) {
-                    return 80;
-                }
+                try {
+                    return Integer.parseInt(host.substring(colonIndex + 1));
+                } catch (NumberFormatException ignore) {}
             }
+            if (getRequestScheme().equals("https")) {
+                return 443;
+            } else if (getRequestScheme().equals("http")) {
+                return 80;
+            }
+
         }
         return getDestinationAddress().getPort();
     }
@@ -779,10 +781,10 @@ public final class HttpServerExchange extends AbstractAttachable {
      * @throws IllegalStateException If this exchange has already been dispatched
      */
     public HttpServerExchange dispatch(final Executor executor, final Runnable runnable) {
-        if (executor != null) {
-            this.dispatchExecutor = executor;
-        }
         if (isInCall()) {
+            if (executor != null) {
+                this.dispatchExecutor = executor;
+            }
             state |= FLAG_DISPATCHED;
             if(anyAreSet(state, FLAG_SHOULD_RESUME_READS | FLAG_SHOULD_RESUME_WRITES)) {
                 throw UndertowMessages.MESSAGES.resumedAndDispatched();
@@ -1119,6 +1121,17 @@ public final class HttpServerExchange extends AbstractAttachable {
      * @param cookie The cookie
      */
     public HttpServerExchange setResponseCookie(final Cookie cookie) {
+        if(getConnection().getUndertowOptions().get(UndertowOptions.ENABLE_RFC6265_COOKIE_VALIDATION, UndertowOptions.DEFAULT_ENABLE_RFC6265_COOKIE_VALIDATION)) {
+            if (cookie.getValue() != null && !cookie.getValue().isEmpty()) {
+                Rfc6265CookieSupport.validateCookieValue(cookie.getValue());
+            }
+            if (cookie.getPath() != null && !cookie.getPath().isEmpty()) {
+                Rfc6265CookieSupport.validatePath(cookie.getPath());
+            }
+            if (cookie.getDomain() != null && !cookie.getDomain().isEmpty()) {
+                Rfc6265CookieSupport.validateDomain(cookie.getDomain());
+            }
+        }
         if (responseCookies == null) {
             responseCookies = new TreeMap<>(); //hashmap is slow to allocate in JDK7
         }
@@ -1568,7 +1581,7 @@ public final class HttpServerExchange extends AbstractAttachable {
                         if (listener.handleDefaultResponse(this)) {
                             return this;
                         }
-                    } catch (Exception e) {
+                    } catch (Throwable e) {
                         UndertowLogger.REQUEST_LOGGER.debug("Exception running default response listener", e);
                     }
                 }
@@ -1587,6 +1600,9 @@ public final class HttpServerExchange extends AbstractAttachable {
             } catch (IOException e) {
                 UndertowLogger.REQUEST_IO_LOGGER.ioException(e);
                 IoUtils.safeClose(connection);
+            } catch (Throwable t) {
+                UndertowLogger.REQUEST_IO_LOGGER.handleUnexpectedFailure(t);
+                IoUtils.safeClose(connection);
             }
         }
 
@@ -1639,8 +1655,12 @@ public final class HttpServerExchange extends AbstractAttachable {
                     } else if (read == -1) {
                         break;
                     }
-                } catch (IOException e) {
-                    UndertowLogger.REQUEST_IO_LOGGER.ioException(e);
+                } catch (Throwable t) {
+                    if (t instanceof IOException) {
+                        UndertowLogger.REQUEST_IO_LOGGER.ioException((IOException) t);
+                    } else {
+                        UndertowLogger.REQUEST_IO_LOGGER.handleUnexpectedFailure(t);
+                    }
                     invokeExchangeCompleteListeners();
                     IoUtils.safeClose(connection);
                     return this;
@@ -1678,11 +1698,15 @@ public final class HttpServerExchange extends AbstractAttachable {
                             public void handleEvent(final StreamSinkChannel channel) {
                                 channel.suspendWrites();
                                 channel.getWriteSetter().set(null);
+                                //defensive programming, should never happen
+                                if (anyAreClear(state, FLAG_RESPONSE_TERMINATED)) {
+                                    UndertowLogger.ROOT_LOGGER.responseWasNotTerminated(connection, HttpServerExchange.this);
+                                    IoUtils.safeClose(connection);
+                                }
                             }
                         }, new ChannelExceptionHandler<Channel>() {
                             @Override
                             public void handleException(final Channel channel, final IOException exception) {
-
                                 //make sure the listeners have been invoked
                                 invokeExchangeCompleteListeners();
                                 UndertowLogger.REQUEST_LOGGER.debug("Exception ending request", exception);
@@ -1691,9 +1715,19 @@ public final class HttpServerExchange extends AbstractAttachable {
                         }
                 ));
                 responseChannel.resumeWrites();
+            } else {
+                //defensive programming, should never happen
+                if (anyAreClear(state, FLAG_RESPONSE_TERMINATED)) {
+                    UndertowLogger.ROOT_LOGGER.responseWasNotTerminated(connection, this);
+                    IoUtils.safeClose(connection);
+                }
+            }
+        } catch (Throwable t) {
+            if (t instanceof IOException) {
+                UndertowLogger.REQUEST_IO_LOGGER.ioException((IOException) t);
+            } else {
+                UndertowLogger.REQUEST_IO_LOGGER.handleUnexpectedFailure(t);
             }
-        } catch (IOException e) {
-            UndertowLogger.REQUEST_IO_LOGGER.ioException(e);
             invokeExchangeCompleteListeners();
 
             IoUtils.safeClose(connection);
@@ -1801,7 +1835,6 @@ public final class HttpServerExchange extends AbstractAttachable {
             requestChannel.runResume();
             ret = true;
         }
-        state &= ~(FLAG_SHOULD_RESUME_READS | FLAG_SHOULD_RESUME_WRITES);
         return ret;
     }
 
@@ -1909,6 +1942,12 @@ public final class HttpServerExchange extends AbstractAttachable {
             }
         }
 
+        @Override
+        public void suspendWrites() {
+            state &= ~FLAG_SHOULD_RESUME_WRITES;
+            super.suspendWrites();
+        }
+
         @Override
         public void wakeupWrites() {
             if (isFinished()) {
@@ -1937,8 +1976,10 @@ public final class HttpServerExchange extends AbstractAttachable {
                 } else {
                     if (wakeup) {
                         wakeup = false;
+                        state &= ~FLAG_SHOULD_RESUME_WRITES;
                         delegate.wakeupWrites();
                     } else {
+                        state &= ~FLAG_SHOULD_RESUME_WRITES;
                         delegate.resumeWrites();
                     }
                 }
@@ -2135,6 +2176,7 @@ public final class HttpServerExchange extends AbstractAttachable {
         @Override
         public void suspendReads() {
             readsResumed = false;
+            state &= ~(FLAG_SHOULD_RESUME_READS);
             super.suspendReads();
         }
 
@@ -2304,8 +2346,10 @@ public final class HttpServerExchange extends AbstractAttachable {
                 } else {
                     if (wakeup) {
                         wakeup = false;
+                        state &= ~FLAG_SHOULD_RESUME_READS;
                         delegate.wakeupReads();
                     } else {
+                        state &= ~FLAG_SHOULD_RESUME_READS;
                         delegate.resumeReads();
                     }
                 }
