@@ -18,8 +18,34 @@
 
 package io.undertow.server.handlers.proxy;
 
+import java.io.Closeable;
+import java.io.IOException;
+import java.net.InetSocketAddress;
+import java.net.SocketAddress;
+import java.net.URI;
+import java.net.URISyntaxException;
+import java.nio.channels.Channel;
+import java.nio.charset.StandardCharsets;
+import java.util.ArrayList;
+import java.util.Collections;
+import java.util.HashMap;
+import java.util.List;
+import java.util.Map;
+import java.util.Set;
+import java.util.concurrent.TimeUnit;
+import javax.net.ssl.SSLPeerUnverifiedException;
+import javax.security.cert.CertificateEncodingException;
+import javax.security.cert.X509Certificate;
+
+import org.jboss.logging.Logger;
+import org.xnio.ChannelExceptionHandler;
+import org.xnio.ChannelListener;
+import org.xnio.ChannelListeners;
+import org.xnio.IoUtils;
+import org.xnio.StreamConnection;
+import org.xnio.XnioExecutor;
+import org.xnio.channels.StreamSinkChannel;
 import io.undertow.UndertowLogger;
-import io.undertow.UndertowOptions;
 import io.undertow.attribute.ExchangeAttribute;
 import io.undertow.attribute.ExchangeAttributes;
 import io.undertow.client.ClientCallback;
@@ -32,6 +58,8 @@ import io.undertow.client.ProxiedRequestAttachments;
 import io.undertow.client.PushCallback;
 import io.undertow.io.IoCallback;
 import io.undertow.io.Sender;
+import io.undertow.predicate.IdempotentPredicate;
+import io.undertow.predicate.Predicate;
 import io.undertow.server.ExchangeCompletionListener;
 import io.undertow.server.HandlerWrapper;
 import io.undertow.server.HttpHandler;
@@ -54,36 +82,7 @@ import io.undertow.util.NetworkUtils;
 import io.undertow.util.SameThreadExecutor;
 import io.undertow.util.StatusCodes;
 import io.undertow.util.Transfer;
-import org.jboss.logging.Logger;
-import org.xnio.ChannelExceptionHandler;
-import org.xnio.ChannelListener;
-import org.xnio.ChannelListeners;
-import org.xnio.IoUtils;
-import org.xnio.StreamConnection;
-import org.xnio.XnioExecutor;
-import org.xnio.channels.StreamSinkChannel;
-
-import javax.net.ssl.SSLPeerUnverifiedException;
-import javax.security.cert.CertificateEncodingException;
-import javax.security.cert.X509Certificate;
-import java.io.Closeable;
-import java.io.IOException;
-import java.io.UnsupportedEncodingException;
-import java.net.InetSocketAddress;
-import java.net.SocketAddress;
-import java.net.URI;
-import java.net.URISyntaxException;
-import java.nio.ByteBuffer;
-import java.nio.channels.Channel;
-import java.nio.charset.Charset;
-import java.nio.charset.StandardCharsets;
-import java.util.ArrayList;
-import java.util.Collections;
-import java.util.HashMap;
-import java.util.List;
-import java.util.Map;
-import java.util.Set;
-import java.util.concurrent.TimeUnit;
+import io.undertow.util.WorkerUtils;
 
 /**
  * An HTTP handler which proxies content to a remote server.
@@ -116,9 +115,11 @@ public final class ProxyHandler implements HttpHandler {
 
     private final HttpHandler next;
 
-    private final boolean rewriteHostHeader;
-    private final boolean reuseXForwarded;
-    private final int maxConnectionRetries;
+    private volatile boolean rewriteHostHeader;
+    private volatile boolean reuseXForwarded;
+    private volatile int maxConnectionRetries;
+
+    private final Predicate idempotentRequestPredicate = IdempotentPredicate.INSTANCE;
 
     public ProxyHandler(ProxyClient proxyClient, int maxRequestTime, HttpHandler next) {
         this(proxyClient, maxRequestTime, next, false, false);
@@ -137,8 +138,8 @@ public final class ProxyHandler implements HttpHandler {
       }
 
     /**
-   *  @param proxyClient the client to use to make the proxy call
-   * @param maxRequestTime the maximum amount of time to allow the request to be processed
+     *  @param proxyClient the client to use to make the proxy call
+     * @param maxRequestTime the maximum amount of time to allow the request to be processed
      * @param next the next handler in line
      * @param rewriteHostHeader should the HOST header be rewritten to use the target host of the call.
      * @param reuseXForwarded should any existing X-Forwarded-For header be used or should it be overwritten.
@@ -165,10 +166,21 @@ public final class ProxyHandler implements HttpHandler {
             next.handleRequest(exchange);
             return;
         }
+        if(exchange.isResponseStarted()) {
+            //we can't proxy a request that has already started, this is basically a server configuration error
+            UndertowLogger.REQUEST_LOGGER.cannotProxyStartedRequest(exchange);
+            exchange.setStatusCode(StatusCodes.INTERNAL_SERVER_ERROR);
+            exchange.endExchange();
+            return;
+        }
         final long timeout = maxRequestTime > 0 ? System.currentTimeMillis() + maxRequestTime : 0;
-        final ProxyClientHandler clientHandler = new ProxyClientHandler(exchange, target, timeout, maxConnectionRetries);
+        int maxRetries = maxConnectionRetries;
+        if(target instanceof ProxyClient.MaxRetriesProxyTarget) {
+            maxRetries = Math.max(maxRetries, ((ProxyClient.MaxRetriesProxyTarget) target).getMaxRetries());
+        }
+        final ProxyClientHandler clientHandler = new ProxyClientHandler(exchange, target, timeout, maxRetries, idempotentRequestPredicate);
         if (timeout > 0) {
-            final XnioExecutor.Key key = exchange.getIoThread().executeAfter(new Runnable() {
+            final XnioExecutor.Key key = WorkerUtils.executeAfter(exchange.getIoThread(), new Runnable() {
                 @Override
                 public void run() {
                     clientHandler.cancel(exchange);
@@ -264,13 +276,15 @@ public final class ProxyHandler implements HttpHandler {
         private final long timeout;
         private final int maxRetryAttempts;
         private final HttpServerExchange exchange;
+        private final Predicate idempotentPredicate;
         private ProxyClient.ProxyTarget target;
 
-        ProxyClientHandler(HttpServerExchange exchange, ProxyClient.ProxyTarget target, long timeout, int maxRetryAttempts) {
+        ProxyClientHandler(HttpServerExchange exchange, ProxyClient.ProxyTarget target, long timeout, int maxRetryAttempts, Predicate idempotentPredicate) {
             this.exchange = exchange;
             this.timeout = timeout;
             this.maxRetryAttempts = maxRetryAttempts;
             this.target = target;
+            this.idempotentPredicate = idempotentPredicate;
         }
 
         @Override
@@ -281,7 +295,7 @@ public final class ProxyHandler implements HttpHandler {
         @Override
         public void completed(final HttpServerExchange exchange, final ProxyConnection connection) {
             exchange.putAttachment(CONNECTION, connection);
-            exchange.dispatch(SameThreadExecutor.INSTANCE, new ProxyAction(connection, exchange, requestHeaders, rewriteHostHeader, reuseXForwarded));
+            exchange.dispatch(SameThreadExecutor.INSTANCE, new ProxyAction(connection, exchange, requestHeaders, rewriteHostHeader, reuseXForwarded, exchange.isRequestComplete() ? this : null, idempotentPredicate));
         }
 
         @Override
@@ -344,36 +358,44 @@ public final class ProxyHandler implements HttpHandler {
         private final Map<HttpString, ExchangeAttribute> requestHeaders;
         private final boolean rewriteHostHeader;
         private final boolean reuseXForwarded;
+        private final ProxyClientHandler proxyClientHandler;
+        private final Predicate idempotentPredicate;
 
         ProxyAction(final ProxyConnection clientConnection, final HttpServerExchange exchange, Map<HttpString, ExchangeAttribute> requestHeaders,
-                           boolean rewriteHostHeader, boolean reuseXForwarded) {
+                    boolean rewriteHostHeader, boolean reuseXForwarded, ProxyClientHandler proxyClientHandler, Predicate idempotentPredicate) {
             this.clientConnection = clientConnection;
             this.exchange = exchange;
             this.requestHeaders = requestHeaders;
             this.rewriteHostHeader = rewriteHostHeader;
             this.reuseXForwarded = reuseXForwarded;
+            this.proxyClientHandler = proxyClientHandler;
+            this.idempotentPredicate = idempotentPredicate;
         }
 
         @Override
         public void run() {
             final ClientRequest request = new ClientRequest();
 
-            StringBuilder requestURI = new StringBuilder();
-            if(!clientConnection.getTargetPath().isEmpty() && !clientConnection.getTargetPath().equals("/")) {
-                requestURI.append(clientConnection.getTargetPath());
-            }
             String targetURI = exchange.getRequestURI();
             if(exchange.isHostIncludedInRequestURI()) {
                 int uriPart = targetURI.indexOf("//");
                 if(uriPart != -1) {
-                    uriPart = targetURI.indexOf("/", uriPart);
-                    targetURI = targetURI.substring(uriPart);
+                    uriPart = targetURI.indexOf("/", uriPart + 2);
+                    if(uriPart != -1) {
+                        targetURI = targetURI.substring(uriPart);
+                    }
                 }
             }
 
             if(!exchange.getResolvedPath().isEmpty() && targetURI.startsWith(exchange.getResolvedPath())) {
                 targetURI = targetURI.substring(exchange.getResolvedPath().length());
             }
+
+            StringBuilder requestURI = new StringBuilder();
+            if(!clientConnection.getTargetPath().isEmpty()
+                    && (!clientConnection.getTargetPath().equals("/") || targetURI.isEmpty())) {
+                requestURI.append(clientConnection.getTargetPath());
+            }
             requestURI.append(targetURI);
 
             String qs = exchange.getQueryString();
@@ -406,9 +428,18 @@ public final class ProxyHandler implements HttpHandler {
                     outboundRequestHeaders.put(entry.getKey(), headerValue.replace('\n', ' '));
                 }
             }
-
+            final String remoteHost;
             final SocketAddress address = exchange.getConnection().getPeerAddress();
-            final String remoteHost = (address != null && address instanceof InetSocketAddress) ? ((InetSocketAddress) address).getHostString() : "localhost";
+            if (address != null && address instanceof InetSocketAddress) {
+                remoteHost = ((InetSocketAddress) address).getHostString();
+                if(!((InetSocketAddress) address).isUnresolved()) {
+                    request.putAttachment(ProxiedRequestAttachments.REMOTE_ADDRESS, ((InetSocketAddress) address).getAddress().getHostAddress());
+                }
+            } else {
+                //should never happen, unless this is some form of mock request
+                remoteHost = "localhost";
+            }
+
             request.putAttachment(ProxiedRequestAttachments.REMOTE_HOST, remoteHost);
 
             if (reuseXForwarded && request.getRequestHeaders().contains(Headers.X_FORWARDED_FOR)) {
@@ -497,8 +528,18 @@ public final class ProxyHandler implements HttpHandler {
                 request.getRequestHeaders().put(Headers.X_FORWARDED_HOST, exchange.getRequestHeaders().getFirst(Headers.HOST));
             }
             if(log.isDebugEnabled()) {
-                log.debugf("Sending request %s to target %s for exchange %s", request, remoteHost, exchange);
+                log.debugf("Sending request %s to target %s for exchange %s", request, clientConnection.getConnection().getPeerAddress(), exchange);
             }
+            //handle content
+            //if the frontend is HTTP/2 then we may need to add a Transfer-Encoding header, to indicate to the backend
+            //that there is content
+            if(!request.getRequestHeaders().contains(Headers.TRANSFER_ENCODING) && !request.getRequestHeaders().contains(Headers.CONTENT_LENGTH)) {
+                if(!exchange.isRequestComplete()) {
+                    request.getRequestHeaders().put(Headers.TRANSFER_ENCODING, Headers.CHUNKED.toString());
+                }
+            }
+
+
             clientConnection.getConnection().sendRequest(request, new ClientCallback<ClientExchange>() {
                 @Override
                 public void completed(final ClientExchange result) {
@@ -514,7 +555,7 @@ public final class ProxyHandler implements HttpHandler {
                             @Override
                             public void handleContinue(final ClientExchange clientExchange) {
                                 if(log.isDebugEnabled()) {
-                                    log.debugf("Relieved continue response to request %s to target %s for exchange %s", request, remoteHost, exchange);
+                                    log.debugf("Received continue response to request %s to target %s for exchange %s", request, clientConnection.getConnection().getPeerAddress(), exchange);
                                 }
                                 HttpContinue.sendContinueResponse(exchange, new IoCallback() {
                                     @Override
@@ -552,7 +593,7 @@ public final class ProxyHandler implements HttpHandler {
                                             path = path.substring(0, i);
                                         }
 
-                                        exchange.dispatch(SameThreadExecutor.INSTANCE, new ProxyAction(new ProxyConnection(pushedRequest.getConnection(), path), exchange, requestHeaders, rewriteHostHeader, reuseXForwarded));
+                                        exchange.dispatch(SameThreadExecutor.INSTANCE, new ProxyAction(new ProxyConnection(pushedRequest.getConnection(), path), exchange, requestHeaders, rewriteHostHeader, reuseXForwarded, null, idempotentPredicate));
                                     }
                                 });
                                 return true;
@@ -561,7 +602,7 @@ public final class ProxyHandler implements HttpHandler {
                     }
 
 
-                    result.setResponseListener(new ResponseCallback(exchange));
+                    result.setResponseListener(new ResponseCallback(exchange, proxyClientHandler, idempotentPredicate));
                     final IoExceptionHandler handler = new IoExceptionHandler(exchange, clientConnection.getConnection());
                     if(requiresContinueResponse) {
                         try {
@@ -569,7 +610,7 @@ public final class ProxyHandler implements HttpHandler {
                                 result.getRequestChannel().getWriteSetter().set(ChannelListeners.flushingChannelListener(new ChannelListener<StreamSinkChannel>() {
                                     @Override
                                     public void handleEvent(StreamSinkChannel channel) {
-                                        Transfer.initiateTransfer(exchange.getRequestChannel(), result.getRequestChannel(), ChannelListeners.closingChannelListener(), new HTTPTrailerChannelListener(exchange, result), handler, handler, exchange.getConnection().getByteBufferPool());
+                                        Transfer.initiateTransfer(exchange.getRequestChannel(), result.getRequestChannel(), ChannelListeners.closingChannelListener(), new HTTPTrailerChannelListener(exchange, result, exchange, proxyClientHandler, idempotentPredicate), handler, handler, exchange.getConnection().getByteBufferPool());
 
                                     }
                                 }, handler));
@@ -580,19 +621,18 @@ public final class ProxyHandler implements HttpHandler {
                             handler.handleException(result.getRequestChannel(), e);
                         }
                     }
-                    Transfer.initiateTransfer(exchange.getRequestChannel(), result.getRequestChannel(), ChannelListeners.closingChannelListener(), new HTTPTrailerChannelListener(exchange, result), handler, handler, exchange.getConnection().getByteBufferPool());
+                    HTTPTrailerChannelListener trailerListener = new HTTPTrailerChannelListener(exchange, result, exchange, proxyClientHandler, idempotentPredicate);
+                    if(!exchange.isRequestComplete()) {
+                        Transfer.initiateTransfer(exchange.getRequestChannel(), result.getRequestChannel(), ChannelListeners.closingChannelListener(), trailerListener, handler, handler, exchange.getConnection().getByteBufferPool());
+                    } else {
+                        trailerListener.handleEvent(result.getRequestChannel());
+                    }
 
                 }
 
                 @Override
                 public void failed(IOException e) {
-                    UndertowLogger.PROXY_REQUEST_LOGGER.proxyRequestFailed(exchange.getRequestURI(), e);
-                    if (!exchange.isResponseStarted()) {
-                        exchange.setStatusCode(StatusCodes.SERVICE_UNAVAILABLE);
-                        exchange.endExchange();
-                    } else {
-                        IoUtils.safeClose(exchange.getConnection());
-                    }
+                    handleFailure(exchange, proxyClientHandler, idempotentPredicate, e);
                 }
             });
 
@@ -600,12 +640,28 @@ public final class ProxyHandler implements HttpHandler {
         }
     }
 
+    static void handleFailure(HttpServerExchange exchange, ProxyClientHandler proxyClientHandler, Predicate idempotentRequestPredicate, IOException e) {
+        UndertowLogger.PROXY_REQUEST_LOGGER.proxyRequestFailed(exchange.getRequestURI(), e);
+        if(exchange.isResponseStarted()) {
+            IoUtils.safeClose(exchange.getConnection());
+        } else if(idempotentRequestPredicate.resolve(exchange) && proxyClientHandler != null) {
+            proxyClientHandler.failed(exchange); //this will attempt a retry if configured to do so
+        } else {
+            exchange.setStatusCode(StatusCodes.SERVICE_UNAVAILABLE);
+            exchange.endExchange();
+        }
+    }
+
     private static final class ResponseCallback implements ClientCallback<ClientExchange> {
 
         private final HttpServerExchange exchange;
+        private final ProxyClientHandler proxyClientHandler;
+        private final Predicate idempotentPredicate;
 
-        private ResponseCallback(HttpServerExchange exchange) {
+        private ResponseCallback(HttpServerExchange exchange, ProxyClientHandler proxyClientHandler, Predicate idempotentPredicate) {
             this.exchange = exchange;
+            this.proxyClientHandler = proxyClientHandler;
+            this.idempotentPredicate = idempotentPredicate;
         }
 
         @Override
@@ -645,18 +701,12 @@ public final class ProxyHandler implements HttpHandler {
                 });
             }
             final IoExceptionHandler handler = new IoExceptionHandler(exchange, result.getConnection());
-            Transfer.initiateTransfer(result.getResponseChannel(), exchange.getResponseChannel(), ChannelListeners.closingChannelListener(), new HTTPTrailerChannelListener(result, exchange), handler, handler, exchange.getConnection().getByteBufferPool());
+            Transfer.initiateTransfer(result.getResponseChannel(), exchange.getResponseChannel(), ChannelListeners.closingChannelListener(), new HTTPTrailerChannelListener(result, exchange, exchange, proxyClientHandler, idempotentPredicate), handler, handler, exchange.getConnection().getByteBufferPool());
         }
 
         @Override
         public void failed(IOException e) {
-            UndertowLogger.PROXY_REQUEST_LOGGER.proxyRequestFailed(exchange.getRequestURI(), e);
-            if (!exchange.isResponseStarted()) {
-                exchange.setStatusCode(StatusCodes.INTERNAL_SERVER_ERROR);
-                exchange.endExchange();
-            } else {
-                IoUtils.safeClose(exchange.getConnection());
-            }
+            handleFailure(exchange, proxyClientHandler, idempotentPredicate, e);
         }
     }
 
@@ -664,10 +714,16 @@ public final class ProxyHandler implements HttpHandler {
 
         private final Attachable source;
         private final Attachable target;
+        private final HttpServerExchange exchange;
+        private final ProxyClientHandler proxyClientHandler;
+        private final Predicate idempotentPredicate;
 
-        private HTTPTrailerChannelListener(final Attachable source, final Attachable target) {
+        private HTTPTrailerChannelListener(final Attachable source, final Attachable target, HttpServerExchange exchange, ProxyClientHandler proxyClientHandler, Predicate idempotentPredicate) {
             this.source = source;
             this.target = target;
+            this.exchange = exchange;
+            this.proxyClientHandler = proxyClientHandler;
+            this.idempotentPredicate = idempotentPredicate;
         }
 
         @Override
@@ -692,11 +748,9 @@ public final class ProxyHandler implements HttpHandler {
                     channel.shutdownWrites();
                 }
             } catch (IOException e) {
-                UndertowLogger.REQUEST_IO_LOGGER.ioException(e);
-                IoUtils.safeClose(channel);
+                handleFailure(exchange, proxyClientHandler, idempotentPredicate, e);
             } catch (Exception e) {
-                UndertowLogger.REQUEST_IO_LOGGER.ioException(new IOException(e));
-                IoUtils.safeClose(channel);
+                handleFailure(exchange, proxyClientHandler, idempotentPredicate, new IOException(e));
             }
 
         }
@@ -732,10 +786,37 @@ public final class ProxyHandler implements HttpHandler {
         }
     }
 
+    public ProxyHandler setMaxConnectionRetries(int maxConnectionRetries) {
+        this.maxConnectionRetries = maxConnectionRetries;
+        return this;
+    }
+
+    public boolean isRewriteHostHeader() {
+        return rewriteHostHeader;
+    }
+
+    public ProxyHandler setRewriteHostHeader(boolean rewriteHostHeader) {
+        this.rewriteHostHeader = rewriteHostHeader;
+        return this;
+    }
+
+    public boolean isReuseXForwarded() {
+        return reuseXForwarded;
+    }
+
+    public ProxyHandler setReuseXForwarded(boolean reuseXForwarded) {
+        this.reuseXForwarded = reuseXForwarded;
+        return this;
+    }
+
     public int getMaxConnectionRetries() {
         return maxConnectionRetries;
     }
 
+    public Predicate getIdempotentRequestPredicate() {
+        return idempotentRequestPredicate;
+    }
+
     private static final class ClosingExceptionHandler implements ChannelExceptionHandler<Channel> {
 
         private final Closeable[] toClose;
@@ -752,47 +833,6 @@ public final class ProxyHandler implements HttpHandler {
         }
     }
 
-    /**
-     * perform URL encoding
-     * <p/>
-     * TODO: this whole thing is kinda crappy.
-     *
-     * @return
-     */
-    private static String encodeUrlPart(final String part, HttpServerExchange exchange) throws UnsupportedEncodingException {
-        //we need to go through and check part by part that a section does not need encoding
-        StringBuilder sb = null;
-        Charset charset = null;
-        for(int i = 0; i < part.length(); ++i) {
-            char c = part.charAt(i);
-            if((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') || (c >= '0' && c <= '9') ||
-                    c == '.' || c == '-' || c == '*' || c == '_' || c == '/') {
-                if(sb != null) {
-                    sb.append(c);
-                }
-            } else {
-                if(sb == null) {
-                    sb = new StringBuilder(part.substring(0, i));
-                    charset = Charset.forName(exchange.getConnection().getUndertowOptions().get(UndertowOptions.URL_CHARSET, UTF_8));
-                }
-                if(c < 127 && charset.name().equals(UTF_8)) {
-                    //minor optimisation
-                    sb.append('%');
-                    sb.append(Integer.toHexString(c));
-                } else {
-                    ByteBuffer bytes = charset.encode(Character.toString(c));
-                    while (bytes.hasRemaining()) {
-                        byte b = bytes.get();
-                        sb.append('%');
-                        sb.append(Integer.toHexString(b & 0xFF));
-                    }
-                }
-            }
-        }
-
-        return sb == null ? part : sb.toString();
-    }
-
     public static class Builder implements HandlerBuilder {
 
         @Override
@@ -847,17 +887,12 @@ public final class ProxyHandler implements HttpHandler {
 
         @Override
         public HttpHandler wrap(HttpHandler handler) {
-
-            final ProxyClient proxyClient;
-            if (uris.size() == 1) {
-                proxyClient = new SimpleProxyClientProvider(uris.get(0));
-            } else {
-                final LoadBalancingProxyClient loadBalancingProxyClient = new LoadBalancingProxyClient();
-                for (URI url : uris) {
-                    loadBalancingProxyClient.addHost(url);
-                }
-                proxyClient = loadBalancingProxyClient;
+            final LoadBalancingProxyClient loadBalancingProxyClient = new LoadBalancingProxyClient();
+            for (URI url : uris) {
+                loadBalancingProxyClient.addHost(url);
             }
+            final ProxyClient proxyClient = loadBalancingProxyClient;
+
             return new ProxyHandler(proxyClient, -1, handler, rewriteHostHeader, false);
         }
     }
