@@ -20,13 +20,13 @@ package io.undertow.server.protocol.framed;
 
 import io.undertow.UndertowLogger;
 import io.undertow.UndertowMessages;
+import io.undertow.connector.PooledByteBuffer;
 import io.undertow.util.ImmediatePooledByteBuffer;
 import org.xnio.Buffers;
 import org.xnio.ChannelListener;
 import org.xnio.ChannelListeners;
 import org.xnio.IoUtils;
 import org.xnio.Option;
-import io.undertow.connector.PooledByteBuffer;
 import org.xnio.XnioExecutor;
 import org.xnio.XnioIoThread;
 import org.xnio.XnioWorker;
@@ -39,9 +39,9 @@ import java.io.InterruptedIOException;
 import java.nio.ByteBuffer;
 import java.nio.channels.FileChannel;
 import java.util.concurrent.TimeUnit;
+import java.util.concurrent.atomic.AtomicIntegerFieldUpdater;
 
 import static org.xnio.Bits.allAreClear;
-import static org.xnio.Bits.allAreSet;
 import static org.xnio.Bits.anyAreSet;
 
 /**
@@ -98,14 +98,18 @@ public abstract class AbstractFramedStreamSinkChannel<C extends AbstractFramedCh
     private volatile SendFrameHeader header;
     private volatile PooledByteBuffer writeBuffer;
     private volatile PooledByteBuffer body;
-    private volatile PooledByteBuffer trailer;
 
     private static final int STATE_CLOSED = 1;
-    private static final int STATE_WRITES_RESUMED = 1 << 1;
-    private static final int STATE_WRITES_SHUTDOWN = 1 << 2;
-    private static final int STATE_IN_LISTENER_LOOP = 1 << 3;
-    private static final int STATE_FIRST_DATA_WRITTEN = 1 << 4;
+    private static final int STATE_WRITES_SHUTDOWN = 1 << 1;
+    private static final int STATE_FIRST_DATA_WRITTEN = 1 << 2;
+    private static final int STATE_PRE_WRITE_CALLED = 1 << 3;
 
+    private volatile boolean bufferFull;
+    private volatile boolean writesResumed;
+    @SuppressWarnings("unused")
+    private volatile int inListenerLoop;
+
+    private static final AtomicIntegerFieldUpdater<AbstractFramedStreamSinkChannel> inListenerLoopUpdater = AtomicIntegerFieldUpdater.newUpdater(AbstractFramedStreamSinkChannel.class, "inListenerLoop");
 
     protected AbstractFramedStreamSinkChannel(C channel) {
         this.channel = channel;
@@ -121,7 +125,7 @@ public abstract class AbstractFramedStreamSinkChannel<C extends AbstractFramedCh
 
     @Override
     public void suspendWrites() {
-        state &= ~STATE_WRITES_RESUMED;
+        writesResumed = false;
     }
 
     /**
@@ -150,28 +154,22 @@ public abstract class AbstractFramedStreamSinkChannel<C extends AbstractFramedCh
         return null;
     }
 
-    /**
-     * Returns the footer for the current frame.
-     *
-     * @return The footer for the current frame, or null
-     */
-    final ByteBuffer getFrameFooter() {
-        if (trailer == null) {
-            trailer = createFrameFooter();
-            if (trailer == null) {
-                trailer = EMPTY_BYTE_BUFFER;
+    final void preWrite() {
+        synchronized (lock) {
+            if (allAreClear(state, STATE_PRE_WRITE_CALLED)) {
+                state |= STATE_PRE_WRITE_CALLED;
+                body = preWriteTransform(body);
             }
         }
-        return trailer.getBuffer();
     }
 
-    protected PooledByteBuffer createFrameFooter() {
-        return null;
+    protected PooledByteBuffer preWriteTransform(PooledByteBuffer body) {
+        return body;
     }
 
     @Override
     public boolean isWriteResumed() {
-        return anyAreSet(state, STATE_WRITES_RESUMED);
+        return writesResumed;
     }
 
     @Override
@@ -185,18 +183,17 @@ public abstract class AbstractFramedStreamSinkChannel<C extends AbstractFramedCh
     }
 
     protected void resumeWritesInternal(boolean wakeup) {
-        boolean alreadyResumed = anyAreSet(state, STATE_WRITES_RESUMED);
+        boolean alreadyResumed = writesResumed;
         if(!wakeup && alreadyResumed) {
             return;
         }
-        state |= STATE_WRITES_RESUMED;
+        writesResumed = true;
         if(readyForFlush && !wakeup) {
             //we already have data queued to be flushed
             return;
         }
 
-        if (!anyAreSet(state, STATE_IN_LISTENER_LOOP)) {
-            state |= STATE_IN_LISTENER_LOOP;
+        if (inListenerLoopUpdater.compareAndSet(this, 0, 1)) {
             getChannel().runInIoThread(new Runnable() {
 
                 int loopCount = 0;
@@ -208,22 +205,24 @@ public abstract class AbstractFramedStreamSinkChannel<C extends AbstractFramedCh
                         if (listener == null || !isWriteResumed()) {
                             return;
                         }
-                        if(loopCount++ == 100) {
+                        if (loopCount++ == 100) {
                             //should never happen
                             UndertowLogger.ROOT_LOGGER.listenerNotProgressing();
                             IoUtils.safeClose(AbstractFramedStreamSinkChannel.this);
                             return;
                         }
                         ChannelListeners.invokeChannelListener((S) AbstractFramedStreamSinkChannel.this, listener);
-                        //if writes are shutdown or we become active then we stop looping
-                        //we stop when writes are shutdown because we can't flush until we are active
-                        //although we may be flushed as part of a batch
 
-                        if (allAreSet(state, STATE_WRITES_RESUMED) && allAreClear(state, STATE_CLOSED) && !broken && !readyForFlush && !fullyFlushed) {
+                    } finally {
+                        inListenerLoopUpdater.set(AbstractFramedStreamSinkChannel.this, 0);
+                    }
+                    //if writes are shutdown or we become active then we stop looping
+                    //we stop when writes are shutdown because we can't flush until we are active
+                    //although we may be flushed as part of a batch
+                    if (writesResumed && allAreClear(state, STATE_CLOSED) && !broken && !readyForFlush && !fullyFlushed) {
+                        if (inListenerLoopUpdater.compareAndSet(AbstractFramedStreamSinkChannel.this, 0, 1)) {
                             getIoThread().execute(this);
                         }
-                    } finally {
-                        state &= ~STATE_IN_LISTENER_LOOP;
                     }
                 }
             });
@@ -233,26 +232,30 @@ public abstract class AbstractFramedStreamSinkChannel<C extends AbstractFramedCh
 
     @Override
     public void shutdownWrites() throws IOException {
-        if(anyAreSet(state, STATE_WRITES_SHUTDOWN) || broken ) {
-            return;
+        synchronized (lock) {
+            if (anyAreSet(state, STATE_WRITES_SHUTDOWN) || broken) {
+                return;
+            }
+            // Queue prior to shutting down writes, since we might send the write buffer
+            queueFinalFrame();
+            state |= STATE_WRITES_SHUTDOWN;
         }
-        // Queue prior to shutting down writes, since we might send the write buffer
-        queueFinalFrame();
-        state |= STATE_WRITES_SHUTDOWN;
     }
 
     private void queueFinalFrame() throws IOException {
-        if (!readyForFlush && !fullyFlushed && allAreClear(state, STATE_CLOSED)  && !broken && !finalFrameQueued) {
-            if( null == body && null != writeBuffer) {
-                sendWriteBuffer();
-            } else if (null == body) {
-                body = EMPTY_BYTE_BUFFER;
+        synchronized (lock) {
+            if (!readyForFlush && !fullyFlushed && allAreClear(state, STATE_CLOSED) && !broken && !finalFrameQueued) {
+                if (null == body && null != writeBuffer) {
+                    sendWriteBuffer();
+                } else if (null == body) {
+                    body = EMPTY_BYTE_BUFFER;
+                }
+                readyForFlush = true;
+                state |= STATE_FIRST_DATA_WRITTEN;
+                state |= STATE_WRITES_SHUTDOWN; // Mark writes as shutdown as well, since we want that set prior to queueing
+                finalFrameQueued = true;
+                channel.queueFrame((S) this);
             }
-            readyForFlush = true;
-            state |= STATE_FIRST_DATA_WRITTEN;
-            state |= STATE_WRITES_SHUTDOWN; // Mark writes as shutdown as well, since we want that set prior to queueing
-            finalFrameQueued = true;
-            channel.queueFrame((S) this);
         }
     }
 
@@ -347,9 +350,11 @@ public abstract class AbstractFramedStreamSinkChannel<C extends AbstractFramedCh
         if (readyForFlush) {
             return false;
         }
-        if (fullyFlushed) {
-            state |= STATE_CLOSED;
-            return true;
+        synchronized (lock) {
+            if (fullyFlushed) {
+                state |= STATE_CLOSED;
+                return true;
+            }
         }
         if (anyAreSet(state, STATE_WRITES_SHUTDOWN) && !finalFrameQueued) {
             queueFinalFrame();
@@ -431,14 +436,14 @@ public abstract class AbstractFramedStreamSinkChannel<C extends AbstractFramedCh
 
     protected boolean safeToSend() throws IOException {
         int state = this.state;
+        if (anyAreSet(state, STATE_CLOSED) || broken) {
+            throw UndertowMessages.MESSAGES.channelIsClosed();
+        }
         if (readyForFlush) {
             return false; //we can't do anything, we are waiting for a flush
         }
         if( null != this.body) {
-            return false; // already have a pooled buffer
-        }
-        if (anyAreSet(state, STATE_CLOSED) || broken) {
-            throw UndertowMessages.MESSAGES.channelIsClosed();
+            throw UndertowMessages.MESSAGES.bodyIsSetAndNotReadyForFlush();
         }
         return true;
     }
@@ -459,11 +464,14 @@ public abstract class AbstractFramedStreamSinkChannel<C extends AbstractFramedCh
     }
 
     private void handleBufferFull() throws IOException {
-        if (!readyForFlush) {
-            sendWriteBuffer();
-            readyForFlush = true;
-            state |= STATE_FIRST_DATA_WRITTEN;
-            channel.queueFrame((S) this);
+        synchronized (lock) {
+            bufferFull = true;
+            if (!readyForFlush) {
+                sendWriteBuffer();
+                readyForFlush = true;
+                state |= STATE_FIRST_DATA_WRITTEN;
+                channel.queueFrame((S) this);
+            }
         }
     }
 
@@ -509,7 +517,9 @@ public abstract class AbstractFramedStreamSinkChannel<C extends AbstractFramedCh
             return;
         }
         try {
-            state |= STATE_CLOSED;
+            synchronized (lock) {
+                state |= STATE_CLOSED;
+            }
             if(writeBuffer != null) {
                 writeBuffer.close();
                 writeBuffer = null;
@@ -522,10 +532,6 @@ public abstract class AbstractFramedStreamSinkChannel<C extends AbstractFramedCh
                 header.getByteBuffer().close();
                 header = null;
             }
-            if (trailer != null) {
-                trailer.close();
-                trailer = null;
-            }
             if (anyAreSet(state, STATE_FIRST_DATA_WRITTEN)) {
                 channelForciblyClosed();
             }
@@ -586,65 +592,74 @@ public abstract class AbstractFramedStreamSinkChannel<C extends AbstractFramedCh
      * Method that is invoked when a frame has been fully flushed. This method is only invoked by the IO thread
      */
     final void flushComplete() throws IOException {
-        try {
-            int remaining = header.getRemainingInBuffer();
-            boolean finalFrame = finalFrameQueued;
-            boolean channelClosed = finalFrame && remaining == 0 && !header.isAnotherFrameRequired();
-            if(remaining > 0) {
-                body.getBuffer().limit(body.getBuffer().limit() + remaining);
-                if(finalFrame) {
-                    //we clear the final frame flag, as it could not actually be written out
-                    //note that we don't attempt to requeue, as whatever stopped it from being written will likely still
-                    //be an issue
+        synchronized (lock) {
+            try {
+                bufferFull = false;
+                int remaining = header.getRemainingInBuffer();
+                boolean finalFrame = finalFrameQueued;
+                boolean channelClosed = finalFrame && remaining == 0 && !header.isAnotherFrameRequired();
+                if (remaining > 0) {
+                    body.getBuffer().limit(body.getBuffer().limit() + remaining);
+                    if (finalFrame) {
+                        //we clear the final frame flag, as it could not actually be written out
+                        //note that we don't attempt to requeue, as whatever stopped it from being written will likely still
+                        //be an issue
+                        this.finalFrameQueued = false;
+                    }
+                } else if (header.isAnotherFrameRequired()) {
                     this.finalFrameQueued = false;
-                }
-            } else if(header.isAnotherFrameRequired()) {
-                this.finalFrameQueued = false;
-                if(body != null) {
+                    if (body != null) {
+                        body.close();
+                        body = null;
+                        state &= ~STATE_PRE_WRITE_CALLED;
+                    }
+                } else if (body != null) {
                     body.close();
                     body = null;
+                    state &= ~STATE_PRE_WRITE_CALLED;
                 }
-            } else if(body != null){
-                body.close();
-                body = null;
-            }
-            if (channelClosed) {
-                fullyFlushed = true;
-                if(body != null) {
-                    body.close();
+                if (channelClosed) {
+                    fullyFlushed = true;
+                    if (body != null) {
+                        body.close();
+                        body = null;
+                        state &= ~STATE_PRE_WRITE_CALLED;
+                    }
+                } else if (body != null) {
+                    // We still have a body, but since we just flushed, we transfer it to the write buffer.
+                    // This works as long as you call write() again
+
+                    //TODO: this code may not work if the channel has frame level compression and flow control
+                    //we don't have an implementation that needs this yet so it is ok for now
+
+                    body.getBuffer().compact();
+                    writeBuffer = body;
                     body = null;
+                    state &= ~STATE_PRE_WRITE_CALLED;
                 }
-            } else if (body != null) {
-                // We still have a body, but since we just flushed, we transfer it to the write buffer.
-                // This works as long as you call write() again
-                body.getBuffer().compact();
-                writeBuffer = body;
-                body = null;
-            }
 
-            if (header.getByteBuffer() != null) {
-                header.getByteBuffer().close();
-            }
-            trailer.close();
-            header = null;
-            trailer = null;
+                if (header.getByteBuffer() != null) {
+                    header.getByteBuffer().close();
+                }
+                header = null;
 
-            readyForFlush = false;
-            if (isWriteResumed() && !channelClosed) {
-                wakeupWrites();
-            } else if(isWriteResumed()) {
-                //we need to execute the write listener one last time
-                //we need to dispatch it back to the IO thread, so we don't invoke it recursivly
-                ChannelListeners.invokeChannelListener(getIoThread(), (S)this, getWriteListener());
-            }
+                readyForFlush = false;
+                if (isWriteResumed() && !channelClosed) {
+                    wakeupWrites();
+                } else if (isWriteResumed()) {
+                    //we need to execute the write listener one last time
+                    //we need to dispatch it back to the IO thread, so we don't invoke it recursivly
+                    ChannelListeners.invokeChannelListener(getIoThread(), (S) this, getWriteListener());
+                }
 
-            final ChannelListener<? super S> closeListener = this.closeSetter.get();
-            if (channelClosed && closeListener != null) {
-                ChannelListeners.invokeChannelListener(getIoThread(), (S) AbstractFramedStreamSinkChannel.this, closeListener);
+                final ChannelListener<? super S> closeListener = this.closeSetter.get();
+                if (channelClosed && closeListener != null) {
+                    ChannelListeners.invokeChannelListener(getIoThread(), (S) AbstractFramedStreamSinkChannel.this, closeListener);
+                }
+                handleFlushComplete(channelClosed);
+            } finally {
+                wakeupWaiters();
             }
-            handleFlushComplete(channelClosed);
-        } finally {
-            wakeupWaiters();
         }
     }
 
@@ -672,13 +687,11 @@ public abstract class AbstractFramedStreamSinkChannel<C extends AbstractFramedCh
                 ChannelListeners.invokeChannelListener(getIoThread(), (S) this, closeListener);
             }
         } finally {
-            if(header != null && header.getByteBuffer() != null) {
-                header.getByteBuffer().close();
-                header = null;
-            }
-            if(trailer != null) {
-                trailer.close();
-                trailer = null;
+            if(header != null) {
+                if( header.getByteBuffer() != null) {
+                    header.getByteBuffer().close();
+                    header = null;
+                }
             }
             if(body != null) {
                 body.close();
@@ -698,7 +711,11 @@ public abstract class AbstractFramedStreamSinkChannel<C extends AbstractFramedCh
     private void wakeupWaiters() {
         if(waiterCount > 0) {
             synchronized (lock) {
-                lock.notifyAll();
+                // It is possible that waiter count would be updated before gaining the lock, lets check one more
+                // time whether the condition wasn't changed in the meantime.
+                if (waiterCount > 0) {
+                    lock.notifyAll();
+                }
             }
         }
     }
@@ -710,4 +727,8 @@ public abstract class AbstractFramedStreamSinkChannel<C extends AbstractFramedCh
     public boolean isBroken() {
         return broken;
     }
+
+    public boolean isBufferFull() {
+        return bufferFull;
+    }
 }
