@@ -18,28 +18,32 @@
 
 package io.undertow.server.protocol.http;
 
-import java.nio.ByteBuffer;
-
 import io.undertow.UndertowOptions;
+import io.undertow.testutils.category.UnitTest;
 import io.undertow.server.HttpServerExchange;
 import io.undertow.util.HttpString;
 import io.undertow.util.Methods;
 import io.undertow.util.Protocols;
+import io.undertow.util.BadRequestException;
 import org.junit.Assert;
 import org.junit.Test;
+import org.junit.experimental.categories.Category;
 import org.xnio.OptionMap;
 
+import java.nio.ByteBuffer;
+
 /**
  * Tests that the parser can resume when it is given partial input
  *
  * @author Stuart Douglas
  */
+@Category(UnitTest.class)
 public class ParserResumeTestCase {
 
     public static final String DATA = "GET http://www.somehost.net/apath+with+spaces%20and%20I%C3%B1t%C3%ABrn%C3%A2ti%C3%B4n%C3%A0li%C5%BE%C3%A6ti%C3%B8n?key1=value1&key2=I%C3%B1t%C3%ABrn%C3%A2ti%C3%B4n%C3%A0li%C5%BE%C3%A6ti%C3%B8n HTTP/1.1\r\nHost:   www.somehost.net\r\nOtherHeader: some\r\n    value\r\nHostee:another\r\nAccept-garbage:   a\r\n\r\ntttt";
     public static final HttpRequestParser PARSER = HttpRequestParser.instance(OptionMap.create(UndertowOptions.ALLOW_ENCODED_SLASH, true));
 
-    final ParseState context = new ParseState();
+    final ParseState context = new ParseState(10);
     @Test
     public void testMethodSplit() {
         byte[] in = DATA.getBytes();
@@ -53,7 +57,7 @@ public class ParserResumeTestCase {
     }
 
     @Test
-    public void testOneCharacterAtATime() throws HttpRequestParser.BadRequestException {
+    public void testOneCharacterAtATime() throws BadRequestException {
         context.reset();
         byte[] in = DATA.getBytes();
         HttpServerExchange result = new HttpServerExchange(null);
@@ -70,7 +74,7 @@ public class ParserResumeTestCase {
         runAssertions(result);
     }
 
-    private void testResume(final int split, byte[] in) throws HttpRequestParser.BadRequestException {
+    private void testResume(final int split, byte[] in) throws BadRequestException {
         context.reset();
         HttpServerExchange result = new HttpServerExchange(null);
         ByteBuffer buffer = ByteBuffer.wrap(in);
