@@ -256,8 +256,9 @@ public class DefaultServer extends BlockJUnit4ClassRunner {
                 if (!DebuggingSlicePool.BUFFERS.isEmpty()) {
                     try {
                         Thread.sleep(200);
-                        if (!DebuggingSlicePool.BUFFERS.isEmpty()) {
-                            Thread.sleep(2000);
+                        long end = System.currentTimeMillis() + 20000;
+                        while (!DebuggingSlicePool.BUFFERS.isEmpty() && System.currentTimeMillis() < end) {
+                            Thread.sleep(200);
                         }
                     } catch (InterruptedException e) {
                         throw new RuntimeException(e);
@@ -372,7 +373,7 @@ public class DefaultServer extends BlockJUnit4ClassRunner {
                     if (h2) {
                         UndertowLogger.ROOT_LOGGER.error("HTTP2 selected but Netty ALPN was not on the boot class path");
                     }
-                    openListener = new HttpOpenListener(pool, OptionMap.create(UndertowOptions.BUFFER_PIPELINED_DATA, true, UndertowOptions.ENABLE_CONNECTOR_STATISTICS, true));
+                    openListener = new HttpOpenListener(pool, OptionMap.builder().set(UndertowOptions.BUFFER_PIPELINED_DATA, true).set(UndertowOptions.ENABLE_CONNECTOR_STATISTICS, true).set(UndertowOptions.REQUIRE_HOST_HTTP11, true).getMap());
                     acceptListener = ChannelListeners.openListenerAdapter(wrapOpenListener(openListener));
                     if (!proxy) {
                         server = worker.createStreamConnectionServer(new InetSocketAddress(Inet4Address.getByName(getHostAddress(DEFAULT)), getHostPort(DEFAULT)), acceptListener, serverOptions);
@@ -513,22 +514,22 @@ public class DefaultServer extends BlockJUnit4ClassRunner {
         } else {
             StringBuilder sb = new StringBuilder(super.testName(method));
             if (isProxy()) {
-                sb.append("{proxy}");
+                sb.append("[proxy]");
             }
             if (ajp) {
-                sb.append("{ajp}");
+                sb.append("[ajp]");
             }
             if (https) {
-                sb.append("{https}");
+                sb.append("[https]");
             }
             if (h2) {
-                sb.append("{http2}");
+                sb.append("[http2]");
             }
             if (h2c) {
-                sb.append("{http2-clear}");
+                sb.append("[http2-clear]");
             }
             if (h2cUpgrade) {
-                sb.append("{http2-clear-upgrade}");
+                sb.append("[http2-clear-upgrade]");
             }
             return sb.toString();
         }
